// #include "src\arduinoFFTfix.h"
#include "src\RadarService.h"
#include <Wire.h>

#include <avr/sleep.h>

#define ACCADR          0x0E
#define VERSION         0xBF010800


#define PEAK_THRESHOLD  0.5
#define SWEEP_RATE      500E3   //(mHz) //this needs to be changed depending on instalation angle & expected flow
#define DISTANCE_RES    60      //(mm)
#define SWEEP_LENGTH    5000    //(mm) //this needs to be changed depending expected water level range
#define SWEEP_START     500     //(mm) //this needs to be changed depending on installation height
#define SWEEPS          64

float peak_threshold    = PEAK_THRESHOLD;
uint32_t sweep_rate     = SWEEP_RATE;
uint8_t distance_res    = DISTANCE_RES;
uint32_t sweep_length   = SWEEP_LENGTH;
uint8_t sweeps          = SWEEPS;
int32_t sweep_start     = SWEEP_START;

RadarService Radar;

uint16_t velocity;
uint16_t distance;
uint16_t amplitude;
float ampl_f;
uint16_t spread;
uint32_t peaks[4];
uint16_t data[128];
char buf[128];
uint8_t cfgMask = 0;

int16_t indexOf(const char* source, char target, int16_t start = 0);

int16_t indexOf(const char* source, char target, int16_t start) {
  if (source == NULL || target == NULL || start < 0 || strlen(source) <= start || '\0' == target) {
    return -1;
  }

  for (int16_t i = start; source[i] != '\0'; i++) {
    if (target == source[i]) {
      return i;
    }
  }

  return -1;
}

int32_t strToInt32(const char* intStr) {
  int8_t sign = 1;

  char* p0 = (char*)intStr;
  if (*p0 == '-') {
    sign = -1;
    p0++;
  } else if (*p0 == '+') {
    p0++;
  }

  char* p = p0;
  while (*p >= '0' && *p <= '9' && *(p + 1) >= '0' && *(p + 1) <= '9') {
    p++;
  }

  char* p1 = p + 1;

  int32_t value = 0;
  int32_t n10 = 1;
  while (p >= p0) {
    value += ((*p--) - '0') * n10;
    n10 *= 10;
  }

  if (*p1 == 'e' || *p1 == 'E') {
    p1++;
    if (*p1 > '0' && *p1 <= '9') {
      n10 = *p1 - '0';
      while (n10-- > 0) {
        value *= 10;
      }
    }
  }

  if (sign == -1) {
    value = -value;
  }

  return value;
}

// cfg:{l:5300,b:300,p:0.5,r:500000,d:60,s:64}
bool parseConfiguration(const char* cfgStr) {
  bool parseSucceed = false;
  char val[8] = { 0 };

  const char* p = strstr(cfgStr, "cfg:{");
  if (p == NULL) {
    return false;
  }

  p = strstr(cfgStr, "{");
  int8_t off = 0;
  for (int8_t i = indexOf(p, ':'); i != -1; i = indexOf(p, ':', i + 1)) {
    off = indexOf(p + i, ',');
    off = off > 0 ? off : indexOf(p + i, '}');
    memset(val, '\0', sizeof(val));
    strncpy(val, p + i + 1, off - 1);

    switch (p[i - 1]) {
      case 'l':  // weep length
        sweep_length = strToInt32(val);
        cfgMask |= 1;
        break;
      case 'b':  // sweep start
        sweep_start = strToInt32(val);
        cfgMask |= 1 << 1;
        break;
      case 'p':  // peek thredshold
        peak_threshold = atof(val);
        cfgMask |= 1 << 2;
        break;
      case 'r':  // sweep rate
        sweep_rate = strToInt32(val);
        cfgMask |= 1 << 3;
        break;
      case 'd':  // distance resolution
        distance_res = strToInt32(val);
        cfgMask |= 1 << 4;
        break;
      case 's':  // sweep
        sweeps = strToInt32(val);
        cfgMask |= 1 << 5;
        break;
      default:;
    }
  }

  return true;
}

void setup() {

  Serial.begin(9600);
  Serial.println(F("Begin"));

  Serial.flush();

  delay(10);
  Serial.println(F("<cfg>"));
  Serial.flush();
  delay(10);

  bool parseConfigurationSucceed = false;
  Serial.setTimeout(1000);
  if (Serial.find("cfg:{")) {
    memset(buf, '\0', sizeof(buf));
    if (Serial.readBytesUntil('}', buf + 5 /*length of "cfg:{"*/, sizeof(buf)) > 0) {
      strncpy(buf, "cfg:{", 5);
      buf[strlen(buf)] = '}';
      parseConfigurationSucceed = parseConfiguration(buf);
    }
  }

  if (Serial.find("<debug>")) {
    if (parseConfigurationSucceed) {
      // cfg:{l:5300,b:300,p:0.5,r:500000,d:60,s:64}
      Serial.write('{');
      if (cfgMask & 1) {
        Serial.write("l:");
        Serial.print(sweep_length, DEC);
        Serial.write(',');
      }
      if (cfgMask & (1 << 1)) {
        Serial.write("b:");
        Serial.print(sweep_start, DEC);
        Serial.write(',');
      }
      if (cfgMask & (1 << 2)) {
        Serial.write("p:");
        Serial.print(peak_threshold, 2);
        Serial.write(',');
      }
      if (cfgMask & (1 << 3)) {
        Serial.write("r:");
        Serial.print(sweep_rate, DEC);
        Serial.write(',');
      }
      if (cfgMask & (1 << 4)) {
        Serial.write("d:");
        Serial.print(distance_res, DEC);
        Serial.write(',');
      }
      if (cfgMask & (1 << 5)) {
        Serial.write("s:");
        Serial.print(sweeps, DEC);
        Serial.write(',');
      }
      Serial.write('}');
      Serial.flush();
    } else {
      Serial.write("{error}");
    }
  }

  //turn on accelerometer
  pinMode(A1, OUTPUT);
  digitalWrite(A1, HIGH);

  Wire.begin();
  delay(10);
  Radar.power_on();
  Radar.init(peak_threshold, sweep_rate, distance_res, sweep_length, sweeps);
}


void loop() {
  delay(10);
  Serial.write('R');
  uint8_t mode = get_mode();
  if (mode) {
    get_radar(sweep_start, mode);

    if (get_ack('A')) {
      Serial.print(F("vel: "));
      Serial.println(velocity);
      Serial.print(F("dist: "));
      Serial.println(distance);
      Serial.print(F("amp: "));
      Serial.println(ampl_f);
      Serial.print(F("spread: "));
      Serial.println(spread);
      Serial.print(F("angle: "));
      float angle = read_acc_angle();
      Serial.println(angle);
      Serial.print(F("peaks: "));
      Serial.print(peaks[0], HEX);
      Serial.print(F("|"));
      Serial.print(peaks[1], HEX);
      Serial.print(F("|"));
      Serial.print(peaks[2], HEX);
      Serial.print(F("|"));
      Serial.println(peaks[3], HEX);
      Serial.print(F("dist_est: "));
      uint16_t edist = est_dist();
      Serial.println(edist);
      Serial.print(F("a_vel: "));
      Serial.println(velocity/cos(angle*0.01745329));
      Serial.print(F("a_dist_est: "));
      Serial.println(edist/sin(angle*0.01745329));

      Serial.flush();
    }
  }
  delay(10);
  Radar.power_off();
  digitalWrite(A1, LOW);
  Wire.end();
  Serial.end();

  // disable ADC
  ADCSRA = 0;

  set_sleep_mode(SLEEP_MODE_PWR_DOWN);
  noInterrupts();  // timed sequence follows
  sleep_enable();

  // turn off brown-out enable in software
  MCUCR = bit(BODS) | bit(BODSE);
  MCUCR = bit(BODS);
  interrupts();  // guarantees next instruction executed
  sleep_cpu();   // sleep within 3 clock cycles of above;
}

void dump_data(void) {
  while (Serial.available()) {
    Serial.read();
  }
  Serial.print(F("D:"));
  //chunksize 256 bytes;
  const uint16_t chunksize = 256;
  uint16_t datalen = Radar.read_reg(0x83);  //get data length
  uint8_t chunks = (sizeof(uint16_t) * datalen) / chunksize;
  Serial.write(chunks);
  for (uint8_t i = 0; i < chunks + 1; i++) {
    if (not get_ack('A')) { break; }

    Radar.read_buf_chunk(data, i * chunksize);
    Serial.write((uint8_t*)data, chunksize);


    while (Serial.available()) { Serial.read(); }
  }
  Serial.print(F("\n"));
}

uint8_t get_mode(void) {

  uint32_t tik = millis();
  while (millis() - tik < 1000) {
    if (Serial.available()) {
      char chr = Serial.read();
      if (chr == 'P') { return 1; }
      if (chr == 'V') { return 2; }
      if (chr == 'M') { return 3; }
      if (chr == 'D') { return 4; }
      if (chr == 'Q') { return 5; }
      if (chr == 'X') { return 6; }
      if (chr == 'E') { return 7; }
      if (chr == 'T') { return 8; }
    }
  }
  return 0;
}

uint8_t get_ack(char ack_char) {
  uint32_t tik = millis();
  while (millis() - tik < 2000) {
    if (Serial.available()) {
      char chr = Serial.read();
      if (chr == ack_char) { return 1; }
      // if(chr == 'N'){return -1;}
    }
  }
  return 0;
}


void get_radar(uint16_t dist, uint8_t mode) {

  Serial.write(mode);
  if (mode == 1) {
    Radar.measure_sparse(dist);
    Radar.eval_sparse(&velocity, &distance, &amplitude, &spread);
    ampl_f = amplitude;
  } else if (mode == 2) {
    Radar.measure_sparse(dist);
    dump_data();
    Radar.eval_sparse(&velocity, &distance, &amplitude, &spread);
    Radar.measure_envelope(dist);
    Radar.eval_envelope(peaks);
    ampl_f = amplitude;
  } else if (mode == 3) {
    ampl_f = Radar.measure_sparse_av(dist, 6);
    Radar.est_sparse(&velocity, &distance, &amplitude, &spread);
  } else if (mode == 4) {
    Radar.measure_sparse(dist);
    Radar.eval_sparse(&velocity, &distance, &amplitude, &spread);
    Radar.measure_envelope(dist);
    Radar.eval_envelope(peaks);
    ampl_f = amplitude;
  } else if (mode == 5) {
    ampl_f = Radar.measure_sparse_av(dist, 6);
    Radar.est_sparse(&velocity, &distance, &amplitude, &spread);

    Radar.measure_envelope(dist);
    Radar.eval_envelope(peaks);

  } else if (mode == 6) {
    ampl_f = Radar.measure_sparse_av(dist, 6);
    dump_data();
    Radar.est_sparse(&velocity, &distance, &amplitude, &spread);

  } else if (mode == 7) {
    Radar.measure_envelope(dist);
    Radar.eval_envelope(peaks);
  } else if (mode == 8) {
    Radar.firmware_rev();
  }
}


/////////////////accelerometer i2c
uint8_t readreg(uint8_t reg) {
  Wire.beginTransmission(ACCADR);
  Wire.write(reg);
  Wire.endTransmission(false);

  Wire.requestFrom(ACCADR, 1);
  return Wire.read();
}

uint8_t writereg(uint8_t reg, uint8_t val) {
  Wire.beginTransmission(ACCADR);
  Wire.write(reg);
  Wire.write(val);
  Wire.endTransmission();
}

float get_angle() {
  delay(20);
  float xg = (readreg(0x07) << 8) + readreg(0x06);
  float yg = (readreg(0x09) << 8) + readreg(0x08);
  float zg = (readreg(0x0B) << 8) + readreg(0x0A);
  float pg = sqrt(xg * xg + yg * yg);
  float cs = atan2(zg, pg);
  return cs * 57.295f;
}

float read_acc_angle() {
  float dec;

  writereg(0x1B, 0b11000000);

  if (readreg(0x1B) == 0b11000000) {
    dec = get_angle();
  }
  writereg(0x1B, 0b01000000);

  return dec;
}


uint16_t est_dist() {
  // uint16_t delta = INT16_MAX;
  uint16_t est_dist = peaks[0] >> sizeof(uint16_t) * 8;
  // uint16_t est_dist;
  // for(uint8_t i = 0; i < 4; i++){
  // uint16_t dist = peaks[i] >> sizeof(uint16_t)*8;
  // uint16_t amp = peaks[i] & 0xFFFF;

  // uint16_t delta_i = (dist < distance) ? (distance - dist) : (dist - distance);
  // if(delta_i < delta){
  // est_dist = dist;
  // delta = delta_i;
  // }
  // }


  return est_dist;
}

// void sparse_6ave(uint16_t dist){
// const int N = 6;
// float acc_bins[SWEEPS/2];
// uint32_t m_dist = 0;
// for(uint8_t i = 0; i < SWEEPS/2; i++){
// acc_bins[i] = 0;
// }
// for(uint8_t i = 0; i < N; i++){
// Radar.measure_sparse(dist);
// // //debug
// // dump_data();
// // //debug
// // Serial.println();
// distance = Radar.pre_proc_sparse(data,acc_bins);

// // dump_data();

// // Radar.est_sparse(&distance);
// m_dist += distance;
// }




// // Serial.println("F`");

// float mx = 0;
// for(uint8_t j = 0; j < SWEEPS/2; j++){
// float d = acc_bins[j];
// d = sqrt(d);
// acc_bins[j] = d;
// // Serial.println(d);
// mx = d > mx ? d : mx;

// }
// mx /= 3;
// for(uint8_t j = 0; j < SWEEPS/2; j++){
// float d = acc_bins[j];
// acc_bins[j] = d > mx ? d : 0;
// }
// float mass = 0;
// float centre = 0;
// for(int j = 0; j < SWEEPS/2; j++){
// float d = acc_bins[j] ;
// mass +=  d;
// centre += d*j;
// }
// centre = centre/mass;
// velocity = centre *(float)(SWEEP_RATE/SWEEPS) * 2.445e-3;
// distance = DISTANCE_RES*m_dist/N + DISTANCE_RES;
// }
